/*********************************************************************************
*                                                                                *
*   This extender allows a winbatch program to open a window, and wait for       *
*   windows messages to arrive.                                                  *
*                                                                                *
*                                                 - Alan Kreutzer,  January 2003 *               
*********************************************************************************/

#define EXTENDERVERSION 10000          // DLL version number

#define INTERFACELEVEL  2              // extender interface level (DON'T CHANGE!)
#define EXTENDER_FEATURE_LEVEL 1001      // extender feature level (DON'T CHANGE)

#define STRICT                         // only if you prototype carefully
/*---------------------------------------------------------------------------
                   Includes and Defines
  ---------------------------------------------------------------------------*/
#include <windows.h> 

// Special platform-dependent includes
#include "wwwpform.h"
#include "addons.h"


#define RETLONG(numb)   return (lpViperCallback(((lpViperReturn->x=numb)      +(lpViperReturn->inittype=VARTYPE_INT))    ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETSTRING(str)  return (lpViperCallback(((lpViperReturn->lpstr=str)   +(lpViperReturn->inittype=VARTYPE_STRING)) ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETERROR(msg)   return (lpViperCallback(((lpViperReturn->lpstr = msg) +(lpViperReturn->inittype=VARTYPE_STRING)) ?(-1):(-1),                     ViperInterfaceVersion, lpViperStruct, lpViperReturn))


/*---------------------------------------------------------------------------+
!                         Structures / Typedefs                              !
+---------------------------------------------------------------------------*/
typedef struct _GLOBAL
{
     ATOM        WindowClass;     /* "Class" of the Progress Window.        */
     HWND        hWnd;            /* hWnd of the Progress window            */   
     HINSTANCE   hInst;           /* hInst of calling program.              */ 
     char       *WindowName;      /* Points to name of window.              */
     long        WindowThread;    /* Thread ID of window handler thread.    */
     long        Error;           /* Set TRUE when the window is open.      */
     HANDLE      EventSignal;     /* Event handle - signals window event.   */     
     long        UserMessage;     /* Message number we are waiting for.     */
     BOOL        MessageReceived; /* Set TRUE when the message is received. */
     WPARAM      wParam;          /* First word of message info.            */
     LPARAM      lParam;          /* Second word of message info.           */
 } GLOBAL;

/*---------------------------------------------------------------------------+
!                              Function Prototypes                           !
+---------------------------------------------------------------------------*/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved);
void WindowThread(GLOBAL *Global);
LRESULT CALLBACK WindowCallback(HWND hWnd,UINT messg,WPARAM wParam,LPARAM lParam);


/*---------------------------------------------------------------------------+
!                               Global Variables                             !
+---------------------------------------------------------------------------*/
/* Superstitious data field to bypass/ignore bugs in certain versions of
   Windows that cause the beginning of the data area to be overwritten.
   This field should be the first data area defined and linked. */
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE hinstDLL;                    // DLL Instance saved

char ExtenderWindowClassName[] = "zwinExtenderWindowClass";

/*---------------------------------------------------------------------------+
!                 Command Extension Table for this DLL                       !
+---------------------------------------------------------------------------*/
tablestruct CommandExtTable[] =
{
//----------------------------------------------------------------------
// ID   Type       Mask               Params/Value Size Name
//----------------------------------------------------------------------
  {10,  FUNCTION,  0x0000000000000000, 0x00000001L,  0, "zwinVersion"       },  
  {11,  FUNCTION,  0x0000000000000001, 0x00000002L,  0, "zwinCreate"        },  
  {12,  FUNCTION,  0x0000000000000000, 0x00000000L,  0, "zwinCheckMessage"  },  
  {13,  FUNCTION,  0x0000000000000000, 0x00000000L,  0, "zwinGetMessage"    },  
  {14,  FUNCTION,  0x0000000000000000, 0x00000000L,  0, "zwinDestroy"       },  
//--------------------------------------------------------------------------
  {-1,         0,                   0,  0x00000000,  0, ""             }
};

/*===========================================================================+
!                                                                            !
!                                                                            !
!                                  DllMain                                   !
!                                                                            !
!                                                                            !
+===========================================================================*/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{
  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);

  return (TRUE);
}

/*===========================================================================+
!                                                                            !
!                              WILExtenderQuery                              !
!                                                                            !
! WILExtenderQuery is an optional (but recommended) entry point.  It is used !
! to provide helpful information to the WIL DLL and possibly other callers.  !
+===========================================================================*/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  static char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*---------------------------------------------------------------------------+
!  case 1 is used to identify yourself to a caller.                          !
+---------------------------------------------------------------------------*/
    case 1:                            // version information
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }

      return (EXTENDERVERSION);
    }

/*---------------------------------------------------------------------------+
!  case 2 is used to inform the WIL DLL of your preferred interface level.   !
+---------------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }

/*---------------------------------------------------------------------------+
!  Return 0 for any unknown requests, or any requests you don't handle!      !
+---------------------------------------------------------------------------*/
    default:
      return (0);
  }
}



/*===========================================================================+
!                                                                            !
!                                                                            !
!                              WILExtender2                                  !
!                                                                            !
!                                                                            !
!                                                                            !
!      This is the routine that performs the actual extender calls.          !
!                                                                            !
!                                                                            !
+===========================================================================*/

DllExport LONG FAR PASCAL WILExtender2
  (HWND            hWnd,                  // hWnd of parent application
  HINSTANCE        hInst,                  // hInstance of parent application
  LONG             ViperInterfaceVersion,  // WIL interface version
  short            CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK  lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT    lpViperStruct,          // pointer to information structure
  LPLONG           lpLong,                 // pointer to your 32 bits
  LPVIPERVAR       lpvvArg,                // pointer to parameter array
  LONG             nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR       lpViperReturn)          // pointer to return value structure
{
  long    X;
  GLOBAL *Global;
  static char szWork[256];                // some local storage

  Global = (GLOBAL *) *lpLong;            // Address of global data block
  
  switch (CmdId)                          // the "big switch" (cases < 0 are at the bottom)
  {

/*----------------------------------------------------------------------------+
!  Case -1 is called once, when your extender is loaded.  It is the place to  !
!  perform your one-time initialization tasks.  The return value should be    !
!  the number of items in your CommandExtTable (functions and constants).     !
+----------------------------------------------------------------------------*/
    case -1:                           // Initialization
    {
       WNDCLASS Class;      
       
       if (lpViperStruct->dwLevel < EXTENDER_FEATURE_LEVEL)
       {
	      lstrcpy((LPSTR)lpvvArg,"zwin Extender: Need newer version of WIL DLL");
          return (-1);
       }

       Global = (GLOBAL *) GlobalAlloc(GMEM_FIXED, sizeof(GLOBAL));
       if (Global == NULL)
       {
          lstrcpy((LPSTR)lpvvArg,"zwin Extender: Not enough memory.");
          return (-1);
       }
      
       Global->WindowClass = 0;
       Global->hWnd = 0;    
       Global->UserMessage = 0;
       Global->EventSignal = NULL;      
       Global->Error = 0;
       Global->MessageReceived = FALSE; 
       Global->wParam = 0; 
       Global->lParam = 0; 

               
       Class.lpszClassName   = ExtenderWindowClassName;
       Class.hInstance       = hInst;
	   Class.lpfnWndProc     = WindowCallback;
       Class.hCursor         = LoadCursor(NULL, IDC_ARROW);
	   Class.hIcon           = LoadIcon(NULL, IDI_APPLICATION);
	   Class.lpszMenuName    = NULL;
       Class.hbrBackground   = (HBRUSH)(COLOR_WINDOW+1);
       Class.style           = 0;
       Class.cbClsExtra      = 0;
       Class.cbWndExtra      = 0;
        
       Global->WindowClass = RegisterClass(&Class);
	   if(Global->WindowClass == 0) 
	   {
	      DWORD Err;
	      char  ErrText[128];
	      Err = GetLastError();
	      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,Err,0,ErrText,128,NULL);
          lstrcpy((LPSTR) lpvvArg,"zwinExtender: Could not create window class.. ");
          lstrcat((LPSTR) lpvvArg,ErrText);  
          return -1;
      }

       *lpLong = (long) Global;      
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1);  // 1 extra at end
    }

/*--------------------------------------------------------------------------+
!  Case -2 is called repeatedly by the WIL extender.  On each call, one     !
!  CommandTable entry is passed back to the WIL extender.                   !
+--------------------------------------------------------------------------*/
    case -2:                           // Loading table
    {
      LPTABLESTRUCT lpTable;
      UINT          j;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      j = (UINT) nArgCount;

      if (CommandExtTable[j].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[j].ident;
        lpTable->type   = CommandExtTable[j].type;
        lpTable->param1 = CommandExtTable[j].param1;
        lpTable->ulMask = CommandExtTable[j].ulMask;
        lpTable->size   = lstrlen((LPSTR) CommandExtTable[j].delim);
        CharLower(CommandExtTable[j].delim);
        lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[j].delim);
      }

      return (1);
    }
    
/*--------------------------------------------------------------------------+
!  Case -3 is called during the WIL shutdown process.  If you need to do    !
!  amy cleanup processing, do it here.                                      !
+--------------------------------------------------------------------------*/
    case -3:                           // shutdown
    {
       if (Global->hWnd)
       {
          SendMessage(Global->hWnd,WM_CLOSE,0,0);
       }
       UnregisterClass(ExtenderWindowClassName,hInst);
       CloseHandle(Global->EventSignal);
       GlobalFree(Global);
       return (1);
    }



/*--------------------------------------------------------------------------+
!                            zwinVersion(i:request)                         !
!                                                                           !
! Returns information on this extender                                      !
!   request: 0 = version, 1 = total functions, 2 = total constants          !
+--------------------------------------------------------------------------*/
    case 10:
    {
      LONG nRequest;
      UINT i, uCount, uType;

      nRequest = lpvvArg[0].x;

      if (nRequest == 0)               // extender version
      {
           RETLONG(EXTENDERVERSION);
      }
 
      if (nRequest == 1)               // total functions
        uType = FUNCTION;
      else if (nRequest == 2)          // total constants
        uType = CONSTANT;
      else
      {
         RETERROR("zwinVersion: Invalid request");
      }

      for (i = 0, uCount = 0; CommandExtTable[i].ident != -1; i++)
        if ((UINT) CommandExtTable[i].type == uType)
          uCount++;

      RETLONG(uCount);
    }

/*--------------------------------------------------------------------------+
!     zwinCreate                                                            !
!                                                                           !
!     Creates the window and saves the message number.                      !
!     This is the number of the message we're waiting for.                  !
!                                                                           !
!     Returns the hWnd of the window.                                       !
+--------------------------------------------------------------------------*/
    case 11:
    {
        HANDLE Event;
        long   Timeout;
        long   EndTicks;
        MSG    Msg;
        
        if (Global->hWnd != 0)
        {
           RETERROR("zwinCreate Error: Window is already open");
        }
        
        /*-----------------------------------------------------+
        !   Create a new signal event.                         !
        !   (If there was an old one, close it first.          !
        !   This is so we don't notice any old signals         !
        !   leftover from the last window.)                    !
        +-----------------------------------------------------*/
        if (Global->EventSignal != NULL)
        {
            CloseHandle(Global->EventSignal);
            Global->EventSignal = NULL;
        }
	    Global->EventSignal = CreateEvent(0,TRUE,FALSE,0);
	    if (Global->EventSignal == NULL)
	    {
	       DWORD Err;
	       char  ErrText[128];
	       Err = GetLastError();
	       FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,Err,0,ErrText,128,NULL);
           lstrcpy(szWork,"zwinExtender: Could not create an \"event\"... ");
           lstrcat(szWork,ErrText);
           RETERROR(szWork);
        }
        
        /*------------------------------------------------------------+
        !   Create a thread to open the window.  (We put the          !
        !   window code in a separate event so it can process         !
        !   messages while we go on our merry way doing other stuff.) !
        +------------------------------------------------------------*/
       Global->hInst = hInst;
       Global->WindowName = lpvvArg[0].lpstr;
       Global->UserMessage = lpvvArg[1].x;
       
       if (CreateThread(0,0,(LPTHREAD_START_ROUTINE) WindowThread,Global,0,(LPDWORD) &X) == NULL)
       {
           long Err;
           char ErrText[128];
           Err = GetLastError();
	       FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,Err,0,ErrText,128,NULL);
           lstrcpy(szWork,"zwinCreate: Could not create window.. ");
           lstrcat(szWork,ErrText);
           RETERROR(szWork);
       }
       
       /*-------------------------------------------------------------+
       !   Wait a minute for the window to open.  (The WindowThread   !
       !   will trigger the EventSignal when it has succeeded or      !
       !   failed.)                                                   !
       +-------------------------------------------------------------*/
       Global->WindowThread = X;
       Event = Global->EventSignal;
       EndTicks = (long) GetTickCount() + 60000;  // End time in ticks
       while (1)
       {            
           Timeout = EndTicks - (long) GetTickCount();
           if (Timeout < 1) Timeout = 1;
           X = MsgWaitForMultipleObjects(1,&Event,0,60000,255);               
           while(PeekMessage(&Msg,NULL,0,0,PM_REMOVE))
             DispatchMessage(&Msg);
           if (X == WAIT_TIMEOUT)
           {
              RETERROR("zwinCreate: Unexpected error.  Window was not created.");
           }
           if (X == 0) break;
       }
              
       if (Global->hWnd != 0)
       {    
           RETLONG((long)Global->hWnd);
       }
       else
       {
           long Err;
           char ErrText[128];
           Err = Global->Error;
	       FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,Err,0,ErrText,128,NULL);
           lstrcpy(szWork,"zwinCreate: Could not create window.. ");
           lstrcat(szWork,ErrText);
           RETERROR(szWork);           
       }
    }
    
/*---------------------------------------------------------------------------+
!              zwinCheckMessage                                              !
!                                                                            !
!    Returns    0 = No messages                                              !
!               1 = Message has arrived.                                     !
!               2 = The window is not open.                                  !
+---------------------------------------------------------------------------*/
    case 12:
    {
        HANDLE Event;
        
        Event = Global->EventSignal;
        
        if (Global->hWnd == 0) 
        {
            RETLONG(2);
        }
        
        if (Global->MessageReceived)
        {
            RETLONG(1);
        }
        
        RETLONG(0);
    }

/*---------------------------------------------------------------------------+
!        zwinGetMessage                                                      !
!                                                                            !
!        Waits until a message arrives  (or the window is closed.)           !
!                                                                            !
!        Returns "" if the window is closed, or  "wparam:lparam"             !
+---------------------------------------------------------------------------*/
    case 13:
    {
       HANDLE Event;
       MSG    Msg;
       
       Event = Global->EventSignal;
       while (1)
       {       
           if (Global->hWnd == 0)   // Window was closed?
           {
              CloseHandle(Event);
              Global->EventSignal = NULL;
              RETSTRING("");
           }
           
           if (Global->MessageReceived)
           {
               wsprintf(szWork,"%d:%d",Global->wParam,Global->lParam);
               Global->MessageReceived = FALSE;
               ResetEvent(Global->EventSignal);
               RETSTRING(szWork);
           }
           
           X = MsgWaitForMultipleObjects(1,&Event,0,10000,255);   
           while(PeekMessage(&Msg,NULL,0,0,PM_REMOVE))
             DispatchMessage(&Msg);
       }    
    }

/*---------------------------------------------------------------------------+
!               zwinDestroy;                                                 !
+---------------------------------------------------------------------------*/
    case 14:
    {
      
       if (Global->hWnd != 0)
       {
            SendMessage(Global->hWnd,WM_CLOSE,0,0);
       }
       
       if (Global->EventSignal != NULL)
       {
            CloseHandle(Global->EventSignal);
            Global->EventSignal = NULL;
       }
       RETLONG(0);
    }
/*---------------------------------------------------------------------------+
!                                  default case                              !
+---------------------------------------------------------------------------*/
    default:                           // we shouldn't get here
    {
       RETERROR("zwinCreate Error: Invalid Function Call");
    }
  }
                                     // end of "big switch"
}                                  
                                 
/*****************************************************************************
*                                                                            *
*    This function is created as a thread to process window messages.        *
*    (We made it a separate thread because we want to process messages       *
*    while the main thread is blocked doing an FTP function.                 *
*                                                                            *
*****************************************************************************/
void WindowThread(GLOBAL *Global)
{
    HWND       hWnd;
    HINSTANCE  hInst;
    char      *WindowName;
    MSG        Msg;
    long       Sts;

    hInst = Global->hInst;
    WindowName = Global->WindowName;
    
    Global->MessageReceived = FALSE;
    Global->wParam = 0;
    Global->lParam = 0;
    Global->Error = 0;
    
    hWnd = CreateWindow(ExtenderWindowClassName,
	               WindowName,
                   WS_OVERLAPPEDWINDOW |WS_ICONIC,
                   20,                      // x
                   20,                      // y
                   100,                     // x size
		           50,                      // y size
                   (HWND)   NULL,
                   (HMENU)  NULL,
                   (HANDLE) hInst,
                   (LPSTR)  NULL 
	              );

       if (hWnd == 0) 
       {
           Global->hWnd  = 0;
           Global->Error = GetLastError();
           SetEvent(Global->EventSignal);  
           ExitThread(0);         
       }

    Global->hWnd = hWnd;
    SetWindowLong(hWnd, GWL_USERDATA, (long) Global);     
	Sts = ShowWindow(hWnd, SW_HIDE);
	Sts = UpdateWindow(hWnd);              
	SetEvent(Global->EventSignal);  
	
    while (GetMessage (&Msg, 0, 0, 0))
    {
         TranslateMessage (&Msg) ;
         DispatchMessage  (&Msg) ;
    }
    Global->hWnd = 0;
	ExitThread(1);
}
/**************************************************************************
*                                                                         *
*                                                                         *
*    The callback procecure for the  window.                              *
*                                                                         *
*                                                                         *
**************************************************************************/
LRESULT CALLBACK WindowCallback(HWND hWnd,UINT messg,WPARAM wParam,LPARAM lParam)
{
    GLOBAL *Global;

    Global = (GLOBAL *) GetWindowLong(hWnd, GWL_USERDATA);      
    if (Global != NULL)
    {
        if (messg == Global->UserMessage)
        {
            Global->wParam = wParam;
            Global->lParam = lParam;
            Global->MessageReceived = TRUE;
            SetEvent(Global->EventSignal);
            return 0;
        }
    }
    
	if (messg == WM_DESTROY)
    {
        PostMessage(Global->hWnd,WM_QUIT,0,0);
	    PostQuitMessage(0);
	    return 0;
    }

    return( DefWindowProc( hWnd, messg, wParam, lParam ) );
}
